#!/usr/bin/env python3
import json
import sys
import timing_dbs


def html_for_cell(cellname, cell, html):
    def proc_pin(pin):
        if type(pin) is list:
            return " ".join(pin)
        else:
            return pin

    iopaths = []
    setupholds = []
    widths = []
    for entry in cell:
        if entry["type"] == "IOPath":
            iopaths.append((
                proc_pin(entry["from_pin"]), proc_pin(entry["to_pin"]),
                tuple(entry["rising"]), tuple(entry["rising"])
            ))
        elif entry["type"] == "SetupHold":
            setupholds.append((
                proc_pin(entry["pin"]), proc_pin(entry["clock"]),
                tuple(entry["setup"]), tuple(entry["hold"])
            ))
        elif entry["type"] == "Width":
            widths.append((
                proc_pin(entry["clock"]), tuple(entry["width"])
            ))
        else:
            assert False

    iopaths.sort()
    setupholds.sort()
    widths.sort()
    print("<a name='{}'/>".format(cellname), file=html)
    print("<h2>{}</h2>".format(cellname), file=html)
    if len(iopaths) > 0:
        print("<h3>Propagation Delays</h3>", file=html)
        print("<table width='800'>", file=html)
        print("<tbody>", file=html)
        print("<tr><th rowspan='2'>From Port</th><th rowspan='2'>To Port</th>", file=html)
        print("<th colspan='3'>Low-High Transition (ps)</th><th colspan='3'>High-Low Transition (ps)</th></tr>", file=html)
        print("<tr><th>Min</th><th>Typ</th><th>Max</th><th>Min</th><th>Typ</th><th>Max</th></tr>", file=html)

        trstyle = ""
        for path in iopaths:
            from_pin, to_pin, rising, falling = path
            trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
            print("<tr {}>".format(trstyle), file=html)
            print("<td>{}</td><td>{}</td>".format(from_pin, to_pin), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(rising[0], rising[1], rising[2]), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(falling[0], falling[1], falling[2]), file=html)
            print("</tr>", file=html)
        print("</tbody>", file=html)
        print("</table>", file=html)
    if len(setupholds) > 0:
        print("<h3>Setup/Hold Checks</h3>", file=html)
        print("<table width='800'>", file=html)
        print("<tbody>", file=html)
        print("<tr><th rowspan='2'>From Port</th><th rowspan='2'>To Clock</th>", file=html)
        print("<th colspan='3'>Setup (ps)</th><th colspan='3'>Hold (ps)</th></tr>", file=html)
        print("<tr><th>Min</th><th>Typ</th><th>Max</th><th>Min</th><th>Typ</th><th>Max</th></tr>", file=html)

        trstyle = ""
        for path in setupholds:
            from_pin, clock, setup, hold = path
            trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
            print("<tr {}>".format(trstyle), file=html)
            print("<td>{}</td><td>{}</td>".format(from_pin, clock), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(setup[0], setup[1], setup[2]), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(hold[0], hold[1], hold[2]), file=html)
            print("</tr>", file=html)
        print("</tbody>", file=html)
        print("</table>", file=html)
    if len(widths) > 0:
        print("<h3>Width Checks</h3>", file=html)
        print("<table width='800'>", file=html)
        print("<tbody>", file=html)
        print("<tr><th rowspan='2'>Clock</th>", file=html)
        print("<th colspan='3'>Width (ps)</th><th colspan='3'>Equiv. Freq (MHz)</th></tr>", file=html)
        print("<tr><th>Min</th><th>Typ</th><th>Max</th><th>Min</th><th>Typ</th><th>Max</th></tr>", file=html)

        trstyle = ""
        for wc in widths:
            clock, width = wc
            trstyle = " bgcolor=\"#dddddd\"" if trstyle == "" else ""
            print("<tr {}>".format(trstyle), file=html)
            print("<td>{}</td>".format(clock), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(width[0], width[1], width[2]), file=html)
            print("<td>{:.0f}</td><td>{:.0f}</td><td>{:.0f}</td>".format(0.5e6/width[2], 0.5e6/width[1], 0.5e6/width[0]), file=html)
            print("</tr>", file=html)
        print("</tbody>", file=html)
        print("</table>", file=html)
    print("<hr/>", file=html)


def make_cell_timing_html(dbfile, family, grade, htmlfile):
    with open(dbfile, 'r') as f:
        db = json.load(f)
    with open(htmlfile, 'w') as html:
        print("<html>", file=html)
        print("<head>", file=html)
        print("<title>{} Speed Grade -{} Cell Timings</title>".format(family, grade), file=html)
        print("</head>", file=html)
        print("<body>", file=html)
        print("<h1>{} Speed Grade -{} Cell Timings</h1>".format(family, grade), file=html)
        print("<h3>Contents</h3>", file=html)
        print("<ul>", file=html)
        for cell in sorted(db.keys()):
            print("<li><a href='#{}'>{}</a></li>".format(cell, cell), file=html)
        print("</ul>", file=html)
        print("<hr/>", file=html)
        for cellname in sorted(db.keys()):
            html_for_cell(cellname, db[cellname], html)
        print("</body>", file=html)
        print("</html>", file=html)


def main(argv):
    if len(argv) < 3:
        print("./cell_html.py grade out.html")
    make_cell_timing_html(timing_dbs.cells_db_path("ECP5", argv[1]), "ECP5", argv[1], argv[2])


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import re
import tiles
import json
lc_input_re = re.compile(r'(J?[ABCDM]|CLK|LSR|CE)\d')
lc_output_re = re.compile(r'J?[FQ]\d')

def get_span(wire):
    if (wire.startswith("H") or wire.startswith("V")) and wire[1:3].isdigit():
        return "span" + wire[2] + wire[0].lower() + wire[3].lower()
    return None


def is_denorm(wire):
    if (wire.startswith("H06") or wire.startswith("V06")) and not wire.endswith("03"):
        return True
    if (wire.startswith("H02") or wire.startswith("V02")) and not wire.endswith("01"):
        return True
    return False


def get_distance(a, b):
    ra, ca = tiles.pos_from_name(a, (126, 95), 0, 0)
    rb, cb = tiles.pos_from_name(b, (126, 95), 0, 0)
    return abs(ra-rb) + abs(ca-cb)


def format_rel(a, b):
    ra, ca = tiles.pos_from_name(a, (126, 95), 0, 0)
    rb, cb = tiles.pos_from_name(b, (126, 95), 0, 0)
    rel = ""
    if rb < ra:
        rel += "n{}".format(ra-rb)
    elif rb > ra:
        rel += "s{}".format(rb-ra)

    if cb < ca:
        rel += "w{}".format(ca-cb)
    elif cb > ca:
        rel += "e{}".format(cb-ca)

    if rel != "":
        rel = "_" + rel
    return rel


def get_pip_class(source, sink):
    source_loc, source_base = source.split("_", 1)
    sink_loc, sink_base = sink.split("_", 1)
    if is_denorm(source_base) or is_denorm(sink_base):
        return None
    if source_base.endswith("_SLICE") or source_base.startswith("MUX") or sink_base.endswith("_SLICE"):
        return "slice_internal"
    if sink_base.endswith("_EBR") or source_base.endswith("_EBR"):
        return "ebr_internal"
    if "TEST" in sink_base or "TEST" in source_base:
        return None
    if "ALU" in sink_base or "ALU" in source_base or "MULT" in sink_base or "MULT" in source_base or "PRADD" in sink_base:
        return "dsp_internal"
    if lc_input_re.match(sink_base):
        if lc_output_re.match(source_base):
            return "{}_to_{}".format(source_base[:-1].lower(), sink_base[:-1].lower())
        elif get_span(source_base) is not None:
            return get_span(source_base)  + "_to_" + sink_base[:-1].lower() + format_rel(source_loc, sink_loc)
        elif "HPBX" in source_base:
            return "global_to_" + sink_base[:-1].lower()
        else:
            return None
    elif get_span(sink_base) is not None:
        if lc_output_re.match(source_base):
            return source_base[:-1].lower() + "_to_" + get_span(sink_base) + format_rel(source_loc, sink_loc)
        elif get_span(source_base) is not None:
            return  get_span(source_base) + "_to_" + get_span(sink_base) + format_rel(source_loc, sink_loc)
        elif "HPBX" in source_base:
            return "global_to_" + get_span(sink_base)
        elif "BOUNCE" in source_base:
            return None
        else:
            assert False, (source, sink)
    elif source_base.startswith("LSR") and sink_base.startswith("MUXLSR"):
        return "lsr_to_muxlsr"
    else:
        return None

# Force these pip classes to have a 0 fanout adder for a better solution
zero_fanout_classes = {
    "span0hr_to_ce",
    "span0vt_to_lsr",
    "span1he_to_m",
    "span1he_to_m_w1",
    "span1he_to_span2vn_n1",
    "span1he_to_span6he_e3",
    "span1hw_to_m",
    "span1hw_to_m_w1",
    "span1hw_to_span6hw_w4",
    "span1vn_to_span2vn_n1",
    "span1vn_to_span6he_e3",
    "span1vn_to_span6hw_w3",
    "span2he_to_lsr",
    "span2he_to_lsr_e1",
    "span2he_to_m",
    "span2he_to_span2vn_n1",
    "span2he_to_span2vn_n1e1",
    "span2he_to_span6he_e4",
    "span2hw_to_lsr",
    "span2hw_to_lsr_w1",
    "span2hw_to_m",
    "span2hw_to_m_w1",
    "span2hw_to_span2vn_n1",
    "span2hw_to_span2vn_n1w1",
    "span2hw_to_span6hw_w4",
    "span6vs_to_span1vs_s3",
    "span6vn_to_span6he_e3",
    "span6vn_to_span6he_n3e3",
    "span6vn_to_span6hw_w3",
    "span6vs_to_span2vn_s2",
    "span6vs_to_span6he_e3",
    "span6vs_to_span6hw_s3w3",
    "span1he_to_span1hw",
    "span1vn_to_m_s1",
    "span1vs_to_m",
    "span6he_to_span1he_e3",
    "span2vs_to_c",
    "span0hr_to_m",
}


def force_zero_fanout_pip(name):
    return name.startswith("jf") or name.startswith("jq") or "to_j" in name or name.startswith("q_to_") or name == "slice_internal" or name == "ebr_internal" or name in zero_fanout_classes

# Force these pip classes to have zero delay to avoid linear dependence and unstable results

zero_delay_classes = {
    "ebr_internal",
    "f_to_span1he_e1",
    "f_to_span1hw",
    "f_to_span1vs",
    "f_to_span2vn_n1",
    "f_to_span6hw_w3",
    "jf_to_span1he_e1",
    "jf_to_span1hw",
    "jf_to_span2vn_n1",
    "jq_to_span1hw",
    "jq_to_span6hw_w3",
    "lsr_to_muxlsr",
    "q_to_span1hw",
    "q_to_span1vs",
    "q_to_span2vn_n1",
    "q_to_span6he_e3",
    "q_to_span6hw_w3",
    "span0hl_to_c"
    "span0hl_to_ce",
    "span0hl_to_d",
    "span0hl_to_m",
    "span0vt_to_lsr",
    "span1he_to_span2vn_n1",
    "span1he_to_span6he_e3",
    "span1hw_to_span6hw_w4",
    "span1vn_to_span2vn_n1",
    "span1vn_to_span6he_e3",
    "span1vn_to_span6hw_w3",
    "span2he_to_lsr",
    "span2he_to_lsr_e1",
    "span2he_to_span2vn_n1",
    "span2he_to_span2vn_n1e1",
    "span2he_to_span6he_e4",
    "span2hw_to_lsr",
    "span2hw_to_lsr_w1",
    "span2hw_to_span2vn_n1",
    "span2hw_to_span2vn_n1w1",
    "span2hw_to_span6hw_w4",
    "span2vs_to_d",
    "span2vs_to_span0hr",
    "span2vs_to_span0vb",
    "span6vn_to_span1vs_n3",
    "span6vs_to_span1vs_s3",
    "span6vs_to_span2vn_s2",
    "span6vs_to_span6hw_w3",
    "span6vn_to_span6he_e3",
    "span6vn_to_span6he_n3e3",
    "span6vn_to_span6hw_w3",
    "span6vs_to_span2vn_s2",
    "span6vs_to_span6he_e3",
    "span6vs_to_span6hw_s3w3",
    "span1he_to_span1hw",
    "span6he_to_span1he_e3",

}

def force_zero_delay_pip(name):
    return name in zero_delay_classes


import re
import tiles

# REGEXs for global/clock signals

# Globals
global_entry_re = re.compile(r'R\d+C\d+_VPRX(\d){2}00')
global_left_right_re = re.compile(r'R\d+C\d+_HPSX(\d){2}00')
global_up_down_re = re.compile(r'R\d+C\d+_VPTX(\d){2}00')
global_branch_re = re.compile(r'R\d+C\d+_HPBX(\d){2}00')

# High Fanout Secondary Nets
hfsn_entry_re = re.compile(r'R\d+C\d+_VSRX(\d){2}00')
hfsn_left_right_re = re.compile(r'R\d+C\d+_HSSX(\d){2}00')
# L2Rs control bidirectional portion of HFSNs!!
hfsn_l2r_re = re.compile(r'R\d+C\d+_HSSX(\d){2}00_[RL]2[LR]')
hfsn_up_down_re = re.compile(r'R\d+C\d+_VSTX(\d){2}00')
# HSBX(\d){2}00 are fixed connections to HSBX(\d){2}01s.
hfsn_branch_re = re.compile(r'R\d+C\d+_HSBX(\d){2}0[01]')

# Center Mux
# Outputs- entry to DCCs connected to globals (VPRXI -> DCC -> VPRX) *
center_mux_glb_out_re = re.compile(r'R\d+C\d+_VPRXCLKI\d+')
# Outputs- connected to ECLKBRIDGEs *
center_mux_ebrg_out_re = re.compile(r'R\d+C\d+_EBRG(\d){1}CLK(\d){1}')

# Inputs- CIB routing to HFSNs
cib_out_to_hfsn_re = re.compile(r'R\d+C\d+_JSNETCIB([TBRL]|MID)(\d){1}')
# Inputs- CIB routing to globals
cib_out_to_glb_re = re.compile(r'R\d+C\d+_J?PCLKCIB(L[TBRL]Q|MID|VIQ[TBRL])(\d){1}')
# Inputs- CIB routing to ECLKBRIDGEs
cib_out_to_eclk_re = re.compile(r'R\d+C\d+_J?ECLKCIB[TBRL](\d){1}')

# Inputs- Edge clocks dividers
eclk_out_re = re.compile(r'R\d+C\d+_J?[TBRL]CDIV(X(\d){1}|(\d){2})')
# Inputs- PLL
pll_out_re = re.compile(r'R\d+C\d+_J?[LR]PLLCLK\d+')
# Inputs- Clock pads
clock_pin_re = re.compile(r'R\d+C\d+_J?PCLK[TBLR]\d+')

# Part of center-mux but can also be found elsewhere
# DCCs connected to globals *
dcc_sig_re = re.compile(r'R\d+C\d+_J?(CLK[IO]|CE)(\d){1}[TB]?_DCC')

# DCMs- connected to DCCs on globals 6 and 7 *
dcm_sig_re = re.compile(r'R\d+C\d+_J?(CLK(\d){1}_|SEL|DCMOUT)(\d){1}_DCM')

# ECLKBRIDGEs- TODO
eclkbridge_sig_re = re.compile(r'R\d+C\d+_J?(CLK(\d){1}_|SEL|ECSOUT)(\d){1}_ECLKBRIDGECS')

# Oscillator output
osc_clk_re = re.compile(r'R\d+C\d+_J?OSC')

# Soft Error Detection Clock *
sed_clk_re = re.compile(r'R\d+C\d+_J?SEDCLKOUT')

# PLL/DLL Clocks
pll_clk_re = re.compile(r'R\d+C\d+_[TB]ECLK\d')

# PG/INRD/LVDS
pg_re = re.compile(r'R\d+C\d+_PG')
inrd_re = re.compile(r'R\d+C\d+_INRD')
lvds_re = re.compile(r'R\d+C\d+_LVDS')

# DDR
ddrclkpol_re = re.compile(r'R\d+C\d+_DDRCLKPOL')
dqsr90_re = re.compile(r'R\d+C\d+_DQSR90')
dqsw90_re = re.compile(r'R\d+C\d+_DQSW90')

def is_global(wire):
    """Return true if a wire is part of the global clock network"""
    return bool(global_entry_re.match(wire) or
        global_left_right_re.match(wire) or
        global_up_down_re.match(wire) or
        global_branch_re.match(wire) or
        hfsn_entry_re.match(wire) or
        hfsn_left_right_re.match(wire) or
        hfsn_l2r_re.match(wire) or
        hfsn_up_down_re.match(wire) or
        hfsn_branch_re.match(wire) or
        center_mux_glb_out_re.match(wire) or
        center_mux_ebrg_out_re.match(wire) or
        cib_out_to_hfsn_re.match(wire) or
        cib_out_to_glb_re.match(wire) or
        cib_out_to_eclk_re.match(wire) or
        eclk_out_re.match(wire) or
        pll_out_re.match(wire) or
        clock_pin_re.match(wire) or
        dcc_sig_re.match(wire) or
        dcm_sig_re.match(wire) or
        eclkbridge_sig_re.match(wire) or
        osc_clk_re.match(wire) or
        sed_clk_re.match(wire) or
        pll_clk_re.match(wire) or
        pg_re.match(wire) or
        inrd_re.match(wire) or
        lvds_re.match(wire) or
        ddrclkpol_re.match(wire) or
        dqsr90_re.match(wire) or
        dqsw90_re.match(wire))

def handle_family_net(tile, wire, prefix_pos, tile_pos, netname):
    if tile.startswith("CENTER") and global_left_right_re.match(wire):
        if prefix_pos[1] < tile_pos[1]:
            return "L_" + netname
        elif prefix_pos[1] > tile_pos[1]:
            return "R_" + netname
        else:
            assert False, "bad CENTER netname"
    elif "CIB_EBR" in tile and "10K" not in tile and global_up_down_re.match(wire) :
        if prefix_pos[0] < tile_pos[0]:
            return "U_" + netname
        elif prefix_pos[0] > tile_pos[0]:
            return "D_" + netname
        else:
            assert False, "bad CIB_EBR netname"
    elif global_branch_re.match(wire):
        return "BRANCH_" + netname
    elif is_global(wire):
        return "G_" + netname
    else:
        return None



from scipy import optimize
from scipy.sparse import linalg
import numpy
import parse_sdf
import design_pip_classes
import pip_classes
import sys
import math
import json


def solve_pip_delays(ncl, sdf, debug=False):
    path_pip_classes, wire_fanout = design_pip_classes.get_equations(ncl)
    sdf_data = parse_sdf.parse_sdf_file(sdf)
    top_ic = None
    for cell in sdf_data.cells:
        if len(sdf_data.cells[cell].interconnect) != 0:
            top_ic = sdf_data.cells[cell].interconnect
            break
    assert top_ic is not None
    variables = {}
    for path in sorted(path_pip_classes.values()):
        for wire, pipclass in path:
            if (pipclass, "delay") not in variables and not pip_classes.force_zero_delay_pip(pipclass):
                vid = len(variables)
                variables[pipclass, "delay"] = vid
            if not pip_classes.force_zero_fanout_pip(pipclass):
                if (pipclass, "fanout") not in variables:
                    vid = len(variables)
                    variables[pipclass, "fanout"] = vid
    kfid = len(variables)
    data = {}
    corners = "min", "typ", "max"
    i = 0
    A = numpy.zeros((len(path_pip_classes), len(variables)))
    for arc, path in sorted(path_pip_classes.items()):
        for wire, pipclass in path:
            if not pip_classes.force_zero_delay_pip(pipclass):
                A[i, variables[pipclass, "delay"]] += 1
            if not pip_classes.force_zero_fanout_pip(pipclass):
                A[i, variables[pipclass, "fanout"]] += wire_fanout[wire, pipclass]
            if pipclass not in data:
                data[pipclass] = {
                    "delay": [0, 0, 0],
                    "fanout": [0, 0, 0],
                }
        i += 1

    for corner in corners:
        b = numpy.zeros((len(path_pip_classes), ))
        i = 0
        for arc, path in sorted(path_pip_classes.items()):
            src, dest = arc
            srcname = "{}/{}".format(src[0].replace('/', '\\/').replace('.', '\\.'), src[1])
            destname = "{}/{}".format(dest[0].replace('/', '\\/').replace('.', '\\.'), dest[1])
            b[i] = getattr(top_ic[srcname, destname].rising, corner + "v")
            i += 1
        print("Starting least squares solver!")
        x, rnorm = optimize.nnls(A, b)
        for var, j in sorted(variables.items()):
            pipclass, vartype = var
            data[pipclass][vartype][corners.index(corner)] = x[j]
    if debug:
        error = numpy.matmul(A, x) - b
        i = 0
        worst = 0
        sqsum = 0
        for arc, path in sorted(path_pip_classes.items()):
            src, dest = arc
            rel_error = error[i] / b[i]
            print("error {}.{} -> {}.{} = {:.01f} ps ({:.01f} %)".format(src[0], src[1], dest[0], dest[1], error[i], rel_error * 100))
            worst = max(worst, abs(error[i]))
            sqsum += error[i]**2
            i = i + 1
        with open("error.csv", "w") as f:
            print("actual, percent", file=f)
            for i in range(len(path_pip_classes)):
                rel_error = error[i] / b[i]
                print("{}, {}".format(b[i], rel_error * 100), file=f)
        for var, j in sorted(variables.items()):
            print("{}_{} = {:.01f} ps".format(var[0], var[1], x[j]))
        print("Error: {:.01f} ps max, {:.01f} ps RMS".format(worst, math.sqrt(sqsum / len(path_pip_classes))))
    return data

def main():
    data = solve_pip_delays(sys.argv[1], sys.argv[2], debug=True)
    with open("out.json", "w") as f:
        json.dump(data, f, indent=4, sort_keys=True)

if __name__ == "__main__":
    main()



import re
import tiles

import ecp5
import machxo2


# General inter-tile routing
general_routing_re = re.compile('R\d+C\d+_[VH]\d{2}[NESWTLBR]\d{4}')
# CIB signals
cib_signal_re = re.compile('R\d+C\d+_J?[ABCDFMQ]\d')
# CIB clock/control signals
cib_control_re = re.compile('R\d+C\d+_J?(CLK|LSR|CEN|CE)\d')
# CIB bounce signals
cib_bounce_re = re.compile('R\d+C\d+_[NESW]BOUNCE')


def is_cib(wire):
    """Return true if a wire is considered part of the CIB (rather than
       a special function - EBR, DSP, etc)"""
    return bool(general_routing_re.match(wire) or
                cib_signal_re.match(wire) or
                cib_control_re.match(wire) or
                cib_bounce_re.match(wire))


h_wire_regex = re.compile(r'H(\d{2})([EW])(\d{2})(\d{2})')
v_wire_regex = re.compile(r'V(\d{2})([NS])(\d{2})(\d{2})')


def handle_edge_name(chip_size, tile_pos, wire_pos, netname, row_bias, col_bias):
    """
    At the edges of the device, canonical wire names do not follow normal naming conventions, as they
    would mean the nominal position of the wire would be outside the bounds of the chip. Before we add routing to the
    database, we must however normalise these names to what they would be if not near the edges, otherwise there is a
    risk of database conflicts, having multiple names for the same wire.

    chip_size: chip size as tuple (max_row, max_col)
    tile_pos: tile position as tuple (r, c)
    wire_pos: wire nominal position as tuple (r, c)
    netname: wire name without position prefix
    row_bias: If "1", use Lattice's 1-based row indexing
    col_bias: If "1", use Lattice's 1-based column indexing

    Returns a tuple (netname, wire_pos)
    """
    hm = h_wire_regex.match(netname)
    vm = v_wire_regex.match(netname)
    if hm:
        # MachXO2 doesn't appear to have edge nets for span-1s.
        if hm.group(1) == "01":
            if tile_pos[1] == chip_size[1] - 1:
                # H01xyy00 --> x+1, H01xyy01
                assert hm.group(4) == "00"
                return "H01{}{}01".format(hm.group(2), hm.group(3)), (wire_pos[0], wire_pos[1] + 1)
        elif hm.group(1) == "02":
            if tile_pos[1] == (1 - col_bias):
                # H02E0002 --> x-1, H02E0001
                # H02W0000 --> x-1, H02W00001
                if hm.group(2) == "E" and wire_pos[1] == (1 - col_bias) and hm.group(4) == "02":
                    return "H02E{}01".format(hm.group(3)), (wire_pos[0], wire_pos[1] - 1)
                elif hm.group(2) == "W" and wire_pos[1] == (1 - col_bias) and hm.group(4) == "00":
                    return "H02W{}01".format(hm.group(3)), (wire_pos[0], wire_pos[1] - 1)
            elif tile_pos[1] == (chip_size[1] - (1 - col_bias)):
                # H02E0000 --> x+1, H02E0001
                # H02W0002 --> x+1, H02W00001
                if hm.group(2) == "E" and wire_pos[1] == (chip_size[1] - (1 - col_bias)) and hm.group(4) == "00":
                    return "H02E{}01".format(hm.group(3)), (wire_pos[0], wire_pos[1] + 1)
                elif hm.group(2) == "W" and wire_pos[1] == (chip_size[1] - (1 - col_bias)) and hm.group(4) == "02":
                    return "H02W{}01".format(hm.group(3)), (wire_pos[0], wire_pos[1] + 1)
        # Bias only affects columns... span-6s seem to work fine without
        # a correction.
        elif hm.group(1) == "06":
            if tile_pos[1] <= 5:
                # x-2, H06W0302 --> x-3, H06W0303
                # x-2, H06E0004 --> x-3, H06E0003
                # x-1, H06W0301 --> x-3, H06W0303
                # x-1, H06E0305 --> x-3, H06E0303
                if hm.group(2) == "W":
                    return "H06W{}03".format(hm.group(3)), (wire_pos[0], wire_pos[1] - (3 - int(hm.group(4))))
                elif hm.group(2) == "E":
                    return "H06E{}03".format(hm.group(3)), (wire_pos[0], wire_pos[1] - (int(hm.group(4)) - 3))
            if tile_pos[1] >= (chip_size[1] - 5):
                # x+2, H06W0304 --> x+3, H06W0303
                # x+2, H06E0302 --> x+3, H06E0303
                if hm.group(2) == "W":
                    return "H06W{}03".format(hm.group(3)), (wire_pos[0], wire_pos[1] + (int(hm.group(4)) - 3))
                elif hm.group(2) == "E":
                    return "H06E{}03".format(hm.group(3)), (wire_pos[0], wire_pos[1] + (3 - int(hm.group(4))))
        else:
            assert False
    if vm:
        if vm.group(1) == "01":
            if tile_pos[0] == 1:
                # V01N000 --> y-1, V01N0001
                if wire_pos[0] == 1 and vm.group(2) == "N" and vm.group(4) == "00":
                    return "V01{}{}01".format(vm.group(2), vm.group(3)), (wire_pos[0] - 1, wire_pos[1])
                if wire_pos[0] == 1 and vm.group(2) == "S" and vm.group(4) == "01":
                    return "V01{}{}00".format(vm.group(2), vm.group(3)), (wire_pos[0] - 1, wire_pos[1])
        elif vm.group(1) == "02":
            if tile_pos[0] == 1:
                # V02S0002 --> y-1, V02S0001
                # V02N0000 --> y-1, V02N0001
                if vm.group(2) == "S" and wire_pos[0] == 1 and vm.group(4) == "02":
                    return "V02S{}01".format(vm.group(3)), (wire_pos[0] - 1, wire_pos[1])
                elif vm.group(2) == "N" and wire_pos[0] == 1 and vm.group(4) == "00":
                    return "V02N{}01".format(vm.group(3)), (wire_pos[0] - 1, wire_pos[1])
            elif tile_pos[0] == (chip_size[0] - 1):
                # V02S0000 --> y+1, V02S0001
                # V02N0002 --> y+1, V02N00001
                if vm.group(2) == "S" and wire_pos[0] == (chip_size[0] - 1) and vm.group(4) == "00":
                    return "V02S{}01".format(vm.group(3)), (wire_pos[0] + 1, wire_pos[1])
                elif vm.group(2) == "N" and wire_pos[0] == (chip_size[0] - 1) and vm.group(4) == "02":
                    return "V02N{}01".format(vm.group(3)), (wire_pos[0] + 1, wire_pos[1])
        elif vm.group(1) == "06":
            if tile_pos[0] <= 5:
                # y-2, V06N0302 --> y-3, H06W0303
                # y-2, V06S0004 --> y-3, V06S0003
                # y-1, V06N0301 --> y-3, V06N0303
                # y-1, V06S0005 --> y-3, V06S0003
                if vm.group(2) == "N":
                    return "V06N{}03".format(vm.group(3)), (wire_pos[0] - (3 - int(vm.group(4))), wire_pos[1])
                elif vm.group(2) == "S":
                    return "V06S{}03".format(vm.group(3)), (wire_pos[0] - (int(vm.group(4)) - 3), wire_pos[1])
            if tile_pos[0] >= (chip_size[0] - 5):
                # y+2, V06N0304 --> y+3, V06N0303
                # y+2, V06S0302 --> x+3, V06S0303
                if vm.group(2) == "N":
                    return "V06N{}03".format(vm.group(3)), (wire_pos[0] + (int(vm.group(4)) - 3), wire_pos[1])
                elif vm.group(2) == "S":
                    return "V06S{}03".format(vm.group(3)), (wire_pos[0] + (3 - int(vm.group(4))), wire_pos[1])
        else:
            assert False
    return netname, wire_pos


def normalise_name(chip_size, tile, wire, family):
    """
    Wire name normalisation for tile wires and fuzzing
    All net names that we have access too are canonical, global names
    These are thus no good for building up a database that is the same for all tiles
    of a given type, as the names will be different in each location.

    Lattice names are of the form R{r}C{c}_{NETNAME}

    Hence, we normalise names in the following way:
     - Global wires have the prefix "G_" added
     - Wires where (r, c) correspond to the current tile have their prefix removed
     - Wires to the left (in TAP_DRIVEs) are given the prefix L, and wires to the right
       are given the prefix R
     - Wires within a DQS group are given the prefix DQSG_
     - Wires within a bank are given the prefix BNK_
     - Other wires are given a relative position prefix using the syntax
       ([NS]\d+)?([EW]\d+)?_
       so a wire whose nominal location is 6 tiles up would be given a prefix N6_
       a wire whose nominal location is 2 tiles down and 1 tile right would be given a prefix
       S2E1_

    TODO: this is more complicated at the edges of the device, where irregular names are used to keep the row and column
    of the nominal position in bounds. Extra logic will be needed to catch and regularise these cases.

    chip_size: chip size as tuple (max_row, max_col)
    tile: name of the relevant tile
    wire: full Lattice name of the wire
    family: Device family to normalise. Affects column indexing (e.g. MachXO2 uses 1-based
          column indexing) and naming of global wires, TAP_DRIVEs, DQS, bank wires,
          etc..

    Returns the normalised netname
    """

    if family == "ECP5":
        def handle_family_net(tile, wire, prefix_pos, tile_pos, netname):
            return ecp5.handle_family_net(tile, wire, prefix_pos, tile_pos, netname)
        row_bias = 0
        col_bias = 0
    elif family.startswith("MachXO"):
        def handle_family_net(tile, wire, prefix_pos, tile_pos, netname):
            return machxo2.handle_family_net(tile, wire, prefix_pos, tile_pos, netname)
        row_bias = 1 if family == "MachXO" else 0
        col_bias = 1
    else:
        raise ValueError("Unknown device family.")

    upos = wire.index("_")
    prefix = wire[:upos]
    prefix_pos = tiles.pos_from_name(prefix, chip_size, row_bias, col_bias)
    tile_pos = tiles.pos_from_name(tile, chip_size, row_bias, col_bias)
    netname = wire[upos+1:]

    family_net = handle_family_net(tile, wire, prefix_pos, tile_pos, netname)
    if family_net:
        return family_net

    netname, prefix_pos = handle_edge_name(chip_size, tile_pos, prefix_pos, netname, row_bias, col_bias)
    if tile_pos == prefix_pos:
        return netname
    else:
        prefix = ""
        if prefix_pos[0] < tile_pos[0]:
            prefix += "N{}".format(tile_pos[0] - prefix_pos[0])
        elif prefix_pos[0] > tile_pos[0]:
            prefix += "S{}".format(prefix_pos[0] - tile_pos[0])
        if prefix_pos[1] > tile_pos[1]:
            prefix += "E{}".format(prefix_pos[1] - tile_pos[1])
        elif prefix_pos[1] < tile_pos[1]:
            prefix += "W{}".format(tile_pos[1] - prefix_pos[1])
        return prefix + "_" + netname


rel_netname_re = re.compile(r'^([NS]\d+)?([EW]\d+)?_.*')

def canonicalise_name(chip_size, tile, wire, row_bias, col_bias):
    """
    Convert a normalised name in a given tile back to a canonical global name
    :param chip_size: chip size as tuple (max_row, max_col)
    :param tile: tilename
    :param wire: normalised netname
    :return: global canonical netname
    """
    if wire.startswith("G_"):
        return wire
    m = rel_netname_re.match(wire)
    tile_pos = tiles.pos_from_name(tile, chip_size, row_bias, col_bias)
    wire_pos = tile_pos
    if m:
        assert len(m.groups()) >= 1
        for g in m.groups():
            if g is not None:
                delta = int(g[1:])
                if g[0] == "N":
                    wire_pos = (wire_pos[0] - delta, wire_pos[1])
                elif g[0] == "S":
                    wire_pos = (wire_pos[0] + delta, wire_pos[1])
                elif g[0] == "W":
                    wire_pos = (wire_pos[0], wire_pos[1] - delta)
                elif g[0] == "E":
                    wire_pos = (wire_pos[0], wire_pos[1] + delta)
        wire = wire.split("_", 1)[1]
    if wire_pos[0] < 0 or wire_pos[0] > chip_size[0] or wire_pos[1] < 0 or wire_pos[1] > chip_size[1]:
        return None #TODO: edge normalisation
    return "R{}C{}_{}".format(wire_pos[0], wire_pos[1], wire)



# Useful functions for constructing nets.
def char_range(c1, c2):
    """Generates the characters from `c1` to `c2`, exclusive."""
    for c in range(ord(c1), ord(c2)):
        yield chr(c)

def net_product(net_list, range_iter):
    return [n.format(i) for i in range_iter for n in net_list]



#!/usr/bin/env python3
import parse_sdf
from os import path
import json
import sys


def include_cell(name, type):
    return type.isupper() and "_" not in type


def rewrite_celltype(name, type):
    return type


def rewrite_pin(name, type, pin):
    return pin


def tupleise(x):
    if type(x) is list:
        return tuple(tupleise(_) for _ in x)
    elif type(x) is tuple:
        return tuple(tupleise(_) for _ in x)
    elif type(x) is dict:
        return "dict", tuple([(k, tupleise(v)) for k, v in sorted(x.items())])
    else:
        return x


def load_database(dbfile):
    if not path.exists(dbfile):
        database = {}
    else:
        database = {}
        with open(dbfile, 'r') as dbf:
            jsondb = json.load(dbf)
        for cell, cdata in jsondb.items():
            database[cell] = set()
            for item in cdata:
                database[cell].add(tupleise(item))
    return database


def make_key(x):
    if type(x) is tuple:
        return ",".join(make_key(_) for _ in x)
    else:
        return str(x)


def save_database(dbfile, database):
    jdb = {}
    for cell, cdata in database.items():
        jcdata = []
        for dtype, dat in sorted(cdata, key=lambda x: make_key(x)):
            assert dtype == "dict"
            jcdata.append({k: v for k, v in dat})
        jdb[cell] = jcdata
    with open(dbfile, 'w') as dbf:
        json.dump(jdb, dbf, indent=4, sort_keys=True)


def delay_tuple(delay):
    return delay.minv, delay.typv, delay.maxv


def add_sdf_to_database(dbfile, sdffile, include_cell_predicate=include_cell, rewrite_cell_func=rewrite_celltype,
                        rewrite_pin_func=rewrite_pin):
    db = load_database(dbfile)
    sdf = parse_sdf.parse_sdf_file(sdffile)
    for instname, cell in sdf.cells.items():
        if not include_cell_predicate(cell.inst, cell.type):
            continue
        celltype = rewrite_cell_func(cell.inst, cell.type)
        if celltype not in db:
            db[celltype] = set()
        for entry in cell.entries:
            if type(entry) is parse_sdf.IOPath:
                db[celltype].add(tupleise(
                    dict(type="IOPath", from_pin=rewrite_pin_func(cell.inst, cell.type, entry.from_pin),
                         to_pin=rewrite_pin_func(cell.inst, cell.type, entry.to_pin), rising=delay_tuple(entry.rising),
                         falling=delay_tuple(entry.falling))))
            elif type(entry) is parse_sdf.SetupHoldCheck:
                db[celltype].add(
                    tupleise(dict(type="SetupHold", pin=rewrite_pin_func(cell.inst, cell.type, entry.pin),
                                  clock=rewrite_pin_func(cell.inst, cell.type, entry.clock),
                                  setup=delay_tuple(entry.setup),
                                  hold=delay_tuple(entry.hold))))
            elif type(entry) is parse_sdf.WidthCheck:
                db[celltype].add(tupleise(dict(type="Width", clock=rewrite_pin_func(cell.inst, cell.type, entry.clock),
                                               width=delay_tuple(entry.width))))
            else:
                assert False
    save_database(dbfile, db)


def main(argv):
    if len(argv) < 3:
        print("Usage: cell_timings.py database.json design.sdf")
        return 2
    add_sdf_to_database(argv[1], argv[2])
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import re
import tiles

# REGEXs for global/clock signals

# Globals including spine inputs, TAP_DRIVE inputs and TAP_DRIVE outputs
global_spine_tap_re = re.compile(r'R\d+C\d+_[HV]P[TLBR]X(\d){2}00')
# CMUX outputs
global_cmux_out_re = re.compile(r'R\d+C\d+_[UL][LR]PCLK\d+')
# CMUX inputs
global_cmux_in_re = re.compile(r'R\d+C\d+_[HV]PF[NESW](\d){2}00')
# Clock pins
clock_pin_re = re.compile(r'R\d+C\d+_J?PCLK[TBLR]\d+')
# PLL global outputs
pll_out_re = re.compile(r'R\d+C\d+_J?[UL][LR][QC]PLL\dCLKO[PS]\d?')

# CIB clock inputs
cib_clk_re = re.compile(r'R\d+C\d+_J?[ULTB][LR][QCM]PCLKCIB\d+')
# Oscillator output
osc_clk_re = re.compile(r'R\d+C\d+_J?OSC')
# Clock dividers
cdivx_clk_re = re.compile(r'R\d+C\d+_J?[UL]CDIVX\d+')
# SED clock output
sed_clk_re = re.compile(r'R\d+C\d+_J?SEDCLKOUT')

# SERDES reference clocks
pcs_clk_re = re.compile(r'R\d+C\d+_J?PCS[AB][TR]XCLK\d')


# DDRDEL delay signals
ddr_delay_re = re.compile(r'R\d+C\d+_[UL][LR]DDRDEL')

# DCC signals
dcc_clk_re = re.compile(r'R\d+C\d+_J?(CLK[IO]|CE)_[BLTR]?DCC(\d+|[BT][LR])')
# DCC inputs
dcc_clki_re = re.compile(r'R\d+C\d+_[BLTR]?DCC(\d+|[BT][LR])CLKI')
# DCS signals
dcs_sig_re = re.compile(r'R\d+C\d+_J?(CLK\d|SEL\d|DCSOUT|MODESEL)_DCS\d')
# DCS clocks
dcs_clk_re = re.compile(r'R\d+C\d+_DCS\d(CLK\d)?')
# Misc. center clocks
center_clk_re = re.compile(r'R\d+C\d+_J?(LE|RE)CLK\d')

# Shared DQS signals
dqs_ssig_re = re.compile(r'R\d+C\d+_(DQS[RW]\d*|(RD|WR)PNTR\d)$')

# Bank edge clocks
bnk_eclk_re = re.compile('R\d+C\d+_BANK\d+(ECLK\d+)')
# CIB ECLK inputs
cib_eclk_re = re.compile(r'R\d+C\d+_J?[ULTB][LR][QCM]ECLKCIB\d+')

brg_eclk_re = re.compile(r'R\d+C(\d+)_JBRGECLK\d+')


def is_global_brgeclk(wire):
    m = brg_eclk_re.match(wire)
    if not m:
        return False
    if m:
        x = int(m.group(1))
        return x > 5 and x < 67

def is_global(wire):
    """Return true if a wire is part of the global clock network"""
    return bool(global_spine_tap_re.match(wire) or
                global_cmux_out_re.match(wire) or
                global_cmux_in_re.match(wire) or
                clock_pin_re.match(wire) or
                pll_out_re.match(wire) or
                cib_clk_re.match(wire) or
                osc_clk_re.match(wire) or
                cdivx_clk_re.match(wire) or
                sed_clk_re.match(wire) or
                ddr_delay_re.match(wire) or
                dcc_clk_re.match(wire) or
                dcc_clki_re.match(wire) or
                dcs_sig_re.match(wire) or
                dcs_clk_re.match(wire) or
                pcs_clk_re.match(wire) or
                center_clk_re.match(wire) or
                cib_eclk_re.match(wire) or
                is_global_brgeclk(wire))

def handle_family_net(tile, wire, prefix_pos, tile_pos, netname):
    if tile.startswith("TAP") and netname.startswith("H"):
        if prefix_pos[1] < tile_pos[1]:
            return "L_" + netname
        elif prefix_pos[1] > tile_pos[1]:
            return "R_" + netname
        else:
            assert False, "bad TAP_DRIVE netname"
    elif is_global(wire):
        return "G_" + netname
    elif dqs_ssig_re.match(wire):
        return "DQSG_" + netname
    elif bnk_eclk_re.match(wire):
        if "ECLK" in tile:
            return "G_" + netname
        else:
            return "BNK_" + bnk_eclk_re.match(wire).group(1)
    elif netname in ("INRD", "LVDS"):
        return "BNK_" + netname
    else:
        return None



import re
import pytrellis


def pos_from_name(tile, chip_size, row_bias, col_bias):
    """
    Extract the tile position as a (row, column) tuple from its name
    """
    size = pytrellis.make_IntPair(chip_size[0],chip_size[1])
    pos = pytrellis.get_row_col_pair_from_chipsize(tile, size, row_bias, col_bias)
    return int(pos.first), int(pos.second)


def type_from_fullname(tile):
    """
    Extract the type from a full tile name (in name:type) format
    """
    return tile.split(":")[1]
